--------------------------------------------------------------------------------
-- A more structured universe for generic deriving
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE DefaultSignatures      #-}
{-# LANGUAGE OverlappingInstances   #-}
{-# LANGUAGE UndecidableInstances   #-}

module Library.Structured ( module Library.Structured, NoPar
                          , module Auxiliary.Metadata ) where


import Auxiliary.Metadata
import Auxiliary.Utils ( NoPar )
import GHC.TypeLits ( Nat )

--------------------------------------------------------------------------------
-- Universe
--------------------------------------------------------------------------------

-- Datatypes
data UData s = UData MetaData (Tree (UCon s))

-- Constructors
data UCon  s = UCon  MetaCons (Tree (USel s))

-- Selectors
data USel  s = USel  MetaSel (UArg s)

-- Constructor arguments
data UArg s = 
             -- Base types and stuff without parameters (e.g. [Int])
               UK1 PRU s

             -- Recursion into parameter containers (e.g. [p], Maybe p)
             | URec SO (s -> s)

             -- The parameter
             | UPar

             -- Composition
             | (s -> s) :.: (UArg s)

-- Parameter, Recursive occurence, or Unknown/other
data PRU = P | R SO | U

-- Self or Other
data SO = S | O

-- Tree structure for constructors and their arguments
data Tree a = Leaf a | Bin (Tree a) (Tree a) | Empty

--------------------------------------------------------------------------------
-- Interpretation
--------------------------------------------------------------------------------

-- We use a poly-kinded data family to prevent having to declare multiple
-- datatypes. In should not be exported, to prevent further instantiation.
data family In :: k -> * -> *

-- Interpretation of datatypes
data instance In (u :: UData a) p where
  D1 :: In l p -> In ('UData cd l) p

-- Interpretation of sums
-- Balanced (or at least tree-like) sums
data instance In (u :: Tree (UCon a)) p where
  -- No constructors
  -- V1 :: In (Empty :: Tree (UCon a)) p
  -- Interpretation of a constructor
  C1 :: In l p -> In (Leaf ('UCon mc l) :: Tree (UCon a)) p
  -- Left/right injection
  L1 :: In l p -> In (Bin l r :: Tree (UCon a)) p
  R1 :: In r p -> In (Bin l r :: Tree (UCon a)) p


-- Interpretation of products
-- Balanced (or at least tree-like) products
data instance In (u :: Tree (USel a)) p where
  -- No arguments
  U1    ::                     In (Empty :: Tree (USel a))   p
  -- Selector field of constructor argument
  S1    :: In x p           -> In (Leaf ('USel ms x))        p
  -- Tupling arguments
  (:*:) :: In l p -> In r p -> In (Bin l r :: Tree (USel a)) p


-- Interpretation of constructor arguments (Structured1)
data instance In (u :: UArg *) p where
  K1   :: { unK1   :: a          } -> In ('UK1 pru a)      p
  Rec  :: { unRec  :: f p        } -> In ('URec so f)      p
  Par  :: { unPar  :: p          } -> In ('UPar :: UArg *) p
  Comp :: { unComp :: t (In f p) } -> In (t :.: f)         p

--------------------------------------------------------------------------------
-- Conversions to/from user datatypes
--------------------------------------------------------------------------------

-- Structured
class Generic (a :: *) where
  type Rep a :: UData *
  type Par a :: *
  type Par a = NoPar

  from :: a -> In (Rep a) (Par a)
  to   :: In (Rep a) (Par a) -> a
  
--------------------------------------------------------------------------------
-- Generic function: representation depth
--------------------------------------------------------------------------------

class GRDepth (a :: k) where
  grdepth :: In a p -> Int

instance (GRDepth cs) => GRDepth ('UData i cs) where
  grdepth (D1 x) = grdepth x

instance GRDepth ('Empty :: Tree (UCon k)) where
  grdepth _ = 0

instance GRDepth ('Leaf a :: Tree (UCon k)) where
  grdepth _ = 1

instance (GRDepth a, GRDepth b) => GRDepth ('Bin a b :: Tree (UCon k)) where
  grdepth (L1 x) = 1 + grdepth x
  grdepth (R1 x) = 1 + grdepth x

class RDepth (a :: *) where
  rdepth :: a -> Int
  default rdepth :: (Generic a, GRDepth (Rep a)) => a -> Int
  rdepth = grdepth . from

--------------------------------------------------------------------------------
-- Generic function: length
--------------------------------------------------------------------------------
{-
class GLengthRep (a :: k) where
  glengthRep :: In a p -> Int

instance (GLengthRep cs) => GLengthRep ('UData i cs) where
  glengthRep (D1 x) = glengthRep x

instance GLengthRep ('Empty :: Tree (UCon k)) where
  glengthRep _ = 0

instance (GLengthRep a) => GLengthRep ('Leaf ('UCon i a) :: Tree (UCon k)) where
  glengthRep (C1 x) = glengthRep x

instance (GLengthRep a, GLengthRep b) => GLengthRep ('Bin a b :: Tree (UCon k)) where
  glengthRep (L1 x) = glengthRep x
  glengthRep (R1 x) = glengthRep x

instance GLengthRep ('Empty :: Tree (USel k)) where
  glengthRep _ = 0

instance (GLengthRep a) => GLengthRep ('Leaf ('USel i a) :: Tree (USel k)) where
  glengthRep (S1 x) = glengthRep x

instance (GLengthRep a, GLengthRep b) => GLengthRep ('Bin a b :: Tree (USel k)) where
  glengthRep (x :*: y) = glengthRep x + glengthRep y

instance (GLength a) => GLengthRep ('UK0 (R so) a) where
  glengthRep (K0 x) = glength x

instance GLengthRep ('UK0 i a :: UArg0 *) where
  glengthRep (K0 x) = 1

class GLength (a :: *) where
  glength :: a -> Int
  default glength :: (Generic0 a, GLengthRep (Rep0 a)) => a -> Int
  glength = glengthRep . from0
-}